/** ECMApp - Simulator for WWII ECM encryption machine 
 * 
 *  This is a translation to javascript, with significant added functionality,
 *  of the ECM Java Applet written by Richard Pekelney
 *  and available at: https://maritime.org/tech/ecmapp.php
 *  Original Java code copyright 1996 Richard Pekelny
 * 
 *  Author: Stephen R. MacMinn
 *  Javascript code copyright 2024 Stephen R. MacMinn
 * 
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8  Oct 96	1.0.0    	RSP 	Origina Java Applet First release.
 * 13 Apr 2024	2.0.0		SRM 	Translated to javascript.  Significant functionality enhancements.
 * 17 Apr 2024	2.1.0		SRM		Added saving and management of rotor sets.
 * 17 May 2024	2.1.2		SRM		Moved loading of csv functions to html file and removed "export"
 * 
 */


const VERSION = '2.1.2';
const VERSION_DATE = '2024-05-17';



/* Globals */

const DECRYPT = true, ENCRYPT = false;
const CSP889 = 0, CSP2900 = 1, CSPNONE = 2;

// Simulator MODE_ states:  
// Simulator will ignore keyboard events when using dialog boxes.

const MODE_RUN = "RUN", MODE_DIALOG = "DIALOG";	
var simulatorMode = MODE_RUN;

// For setting properties of DOM elements.
const PROPERTY_INNERHTML = "innerHTML";
const PROPERTY_VALUE	 = "value";

// Rotor order parameters. 
// These select the order of the rotors and their orientation (normal or
// reversed. Near the end of the war, this was done once each day.
// In a real machine each rotor can be used only once,
// this simulator allows using the same rotor number in multiple positions.
// There are five rotors in each bank. (cipher, control, and index) 
// The string contains an index to the rotor
// wiring table followed by either R (reverse) or N (normal) for each of the rotors
// that make up a bank.  If the parameters are not provided, or are not the correct
// length, defaults are provided.  If an index is not in range, the RotorCage 
// constructor will substitute the first rotor wiring.

var defaultControlOrder = "5N6N7N8N9N"; 
var defaultCipherOrder  = "0N1N2N3N4N"; 
var defaultIndexOrder   = "0N1N2N3N4N"; 

var cipherOrder  = defaultCipherOrder;
var controlOrder = defaultControlOrder; 
var indexOrder   = defaultIndexOrder;

var zeroSwitchId = 'zeroSwitch';      			// zeroize switch
var zs;              							// position of zeroize switch

var machSwitchId = 'machSwitch';      			// type of machine switch
var machType = CSP889; 							// position of machine switch

var masterSwitchId = 'masterSwitch';    		// master switch
var ms;              							// position of the master switch
var oldms = "O";     							// the last position of the master switch.

var countDispId = 'countDisp';    				// counter display
var count = 0;         							// key counter

var paperTapeId = 'paperTape';     				// paper tape output
var keyboardDisplayId = 'keyboardDisplay';		// keyboard display output
var clearKeyboardDisplayId = 'clearKeyboardDisplay';
var tearTapeId  = 'tearTape';      				// clear paper tape button
var encPaperCount = 0; 							// encrypt printing counter, used to create 5 character groups
var copyPaperTapeId = 'copyPaperTape';
var copykeyboardDisplayId = 'copyKeyboardDisplay';
var keySoundOn = true;
    
var smallKeyClass = "smallKey";
var keyClass = "vKeyboard";
 
var cipherDispId  = 'cipherDisp';				// display of cipher rotor positions
var controlDispId = 'controlDisp';				// display of control rotor positions
var indexDispId   = 'indexDisp';   				// display of index rotor positions

var cage;         								// the rotor cage contains the rotor banks and in
                    							// this program, the wiring between rotors.  On the
                    							// real machine the wiring is not in the rotor cage.
                    							// This object encapsulates most of the algorithm.

const storageManager = new ECMStorageManager();

// letterKeySet contains all 26 letters.
	
const letterKeySet = new Set(
	['Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P',
	 'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 
	 'Z', 'X', 'C', 'V', 'B', 'N', 'M' 
	]);
const BLANK_KEY = 'BLANK';	
	
const clickAudio = new Audio();

const aboutDialog = `
	<h2> About ECM Simulator</h2>
	<p>Version: ` + VERSION + `</p>
	<p>Date: ` + VERSION_DATE + `</p>
	<p>Author: Stephen R. MacMinn</p>
	<p>Based on a Java Applet by Rich Pekelny</p>
	<p>Copyrights 1999 and 2024 by Rich Pekelny and Stephen MacMinn</p>
	`;
	
const fileUploadDialog = `
	<h2 style="margin-bottom:1rem;">Add Rotor Set</h2>
	<p>Select a .csv file containing your rotor set definition and give the set a name.</p>
	<p>The file must consist of 10 rows, each with 26 letters.  Each row defines a rotor</p>
	<p>The file must be in valid csv format.</p>
	<input type="text" id="settingsFunction" value="uploadRotorSet" hidden>
	<input type="file" id="uploadRotorsFile" style="margin-top: 1.5rem;" />
	<p><span>Give your Rotor set a name:</span>
	<input type="text" id="rotorSetName" style="margin-top: 1rem;" />
	</p>
	<button id="dialogSubmit" class="clrButton" style="margin-top:10px; margin-bottom:20px; margin-left:auto; margin-right:auto; width:20%">Upload</button>
	`;

var lightbox = document.createElement('div');
lightbox.id = "lightbox";

// Create the outer element for the dialog box
// containing the close icon at the top a div to hold the dialog
// content in the middle, and a message window at the bottom

let dialogFrame = `	
	<div style="width:100%">	
	<img id="dialogClose" onclick='dialogClose()' src="img/close_icon.svg" style="float:right; margin-top:5px; margin-bottom:5px; "width="20" height="20" />
	</div>
	<div id='dialogContent' style="width:100%;"></div>
	<div id="messageWindow" style="text-align:center"></div>
`;

// lightbox is the element that holds the user interaction dialog.
	
lightbox.innerHTML = dialogFrame;
lightbox.style.visibility = "hidden";
lightbox.classList.add("overlay-white");
document.getElementById("ecmBody").appendChild(lightbox);

// Overlay is an element that hides the page body to display the user dialogs.

var overlay = document.createElement('div');	
overlay.id = "overlay";
document.body.appendChild(overlay);	
window.addEventListener('resize', centerOverlay);	// center when the window resizes
centerOverlay();


/* start things up */

init();
action();

/***********************************************************************
 * init() - Initializes ECMapp
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 * Translated to javascript by Stephen R. MacMinn
 *
 * SUMMARY:
 * init() runs each time the web page that contains it loads ECMApp, it then 
 * creates the RotorCage and initializes the positions of the rotors
 * The RotorCage object has the guts of the ECM algorithm.
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    	RSP 	First release.
 * 12 Apr 2024	2.1			SRM		Translated to javascript.
 *      
 ***********************************************************************/
 
 function init() {

    clickAudio.src = "sound/keyClick.mp3";		// Load the key click sound.

	// Create the rotor cage, putting the rotors in the desired order.
	//	
    //                  	Cipher      Control     Index
    cage = new RotorCage(cipherOrder, controlOrder, indexOrder);
    
    // Zeroize all the control and cipher rotors. Zeroizing a rotor is 
    // setting its position to letter 'O'.
    
    cage.zeroize(); 
	
    // By the end of the war the Index rotor order is always set 10,20,30,40,50.
    // The starting positions of index rotors comes from the daily key list and 
    // different positions are used for each classification of message.
    
    cage.setIndexBankPos("00000");
                 	    
    // Make sure the rotor position displays are updated after initialization.
    
    document.getElementById(cipherDispId).value  = cage.cipherBankPosToString();
    document.getElementById(controlDispId).value = cage.controlBankPosToString();
    document.getElementById(indexDispId).value   = cage.indexBankPosToString();	 
}   // end of init  

//
//  Some UI interaction utility functions
//

/** centerOverlay
 *  Center the masking overlay on the page.
 *  This is a function because it needs to be done whenever the user resizes the window.
 *
 */
function centerOverlay() {

  // Get the width and height of the overlay
  const width = overlay.offsetWidth;
  const height = overlay.offsetHeight;

  // Get the width and height of the window
  const windowWidth = window.innerWidth;
  const windowHeight = window.innerHeight;

  // Calculate the top and left positions of the overlay
  const top = (windowHeight - height) / 2;
  const left = (windowWidth - width) / 2;

  // Set the overlay's top and left positions
  overlay.style.top  = top  + 'px';
  overlay.style.left = left + 'px';
  
}  // end of centerOVerlay

// dialogSubmit - Process input from the settings dialog when user presses submit button.

function dialogSubmit() {		
	let func = document.getElementById("settingsFunction").value;
	let rotorSetName;
	let result;
	
	switch(func) {
/*
		case "installRotorSet":
			rotorSetName = document.getElementById("rotorSetName").value;
			result = installRotorSet(rotorSetName);
			break;
*/			
		case "uploadRotorSet":	
			let file = document.getElementById("uploadRotorsFile").files[0];
			rotorSetName = document.getElementById("rotorSetName").value;
			result = processRotorSetUpload(rotorSetName, file);
			break;
			
		default:
	}
}

/* postStatusMessage - Post a message to the messageWindow.
 *
 * @param	status	Array	[0] Boolean with true/false for success/failure.
 * 							[1] A status message.
 * 
 * Author:	S.R. MacMinn
 * 
 * @history	2024-04-11	First version.
 * 
 */
function postMessage(status) {
	
	let success;
	let msgText;
	let span;
	
	if(!status) {
		success = false;
		msgText = "No status passed to postMessage";
	} else {
		success = status[0];
		msgText = status[1];
	}
	
	if(success) {
		span = "<span class=\"statusSuccess\"> ";
	} else {
		span = "<span class=\"statusFailure\"> ";
	}
	msg = span + msgText + "</span>";
	let d = document.getElementById("messageWindow");
	d.innerHTML = msg;

	d.scrollIntoView({ behavior: 'smooth', block: 'end' });
}  // end of postMessage

/* dialogClose - close the settings dialog box
 * Note that dialogClose is called via an onClick parameter in the image tag, so there
 * is no matching EventListener.  This is done because, at initialization time the
 * DOM node does not exist.
 * 
 */

function dialogClose() {
	document.getElementById("overlay").style.visibility = "hidden";
	document.getElementById("dialogContent").innerHTML = "";
	document.getElementById("messageWindow").innerHTML = ""
	document.getElementById("lightbox").style.visibility = "hidden";
	enableKeySound(true);
	simulatorMode = MODE_RUN;
} // end of dialogClose

/* processRotorSetUpload
 * Once user clicks submit on rotor set upload, save the data.
 * Note: postMessage() is called in here rather than in the calling routine
 * because this function is called ashychronously.  
 * 
 */ 

function processRotorSetUpload(name, file) {
	if(!name) {
		return([false, "Please provide a rotor set name"]);
	}
	
	var reader = new FileReader();
	reader.readAsText(file);								
	reader.onerror = function() {
	    console.log(reader.error);
	    postMessage([false, "Error reading file " + file.name]);
	};

	reader.onload = function(event) {
		var csv_string = event.target.result;
		let data = csvParse(csv_string);
		let result = validateRotorSet(data);
		if(!result[0]) {
			postMessage(result);
			return;
		}
		result = storageManager.addRotorSet(name, data);
		postMessage(result);
	}
}

function installRotorSet(rotorSetName) {
	let rotorSet;
	
	if(rotorSetName == Rotor.DEFAULT_ROTOR_SET_NAME) {
		rotorSet = [...defaultWIRING];
	} else {
		rotorSet = storageManager.getRotorSet(rotorSetName);		
	}
	
	if(!rotorSet) {
		return [false, 'Installation failed: ' + rotorSetName + ' not found.'];

	// Replace the rotor WIRING and re-create the rotorCage.
	
	} else {
		Rotor.WIRING = [...rotorSet];
		Rotor.installRotorSet(rotorSetName, rotorSet);
		cage = new RotorCage(cipherOrder, controlOrder, indexOrder);
		cage.zeroize();
	    cage.setIndexBankPos("00000");		
		return [true, 'Rotor set: ' + rotorSetName + ' installed.'];;
	}
}

/* changeRotorOrderHTML
 * Generate HTML to change rotor order.
 *
 */
function changeRotorOrderHTML() {	
	let html = '<h2 style="margin-bottom:1rem; clear:both; text-align:center;">Set Rotor Order</h2>';

		html += `
			<p>
				Set the order of the cipher, control, and index rotors in the rotor cage.
				In a real machine each rotor can be used only once,
				the simulator allows using the same rotor number more than once.
				There are five rotors in each bank. The string contains an index to the rotor
				wiring table followed by either R (reverse) or N (normal) for each of the rotors
		 		that make up a bank, e.g., '1N2N3R4N5R'
		 		Clearing a field will restore the default rotor order for that set.
			</p>
		<input type="text" id="settingsFunction"  value="setRotorOrder" hidden>
		`;
		
		html += '<div style="display:inline; width:50%;"> Cipher order:  </div> <input id="cipherOrder"  type="text"  style="margin-top: 1.5rem; width:50%; letter-spacing: .2em;" value=' + cipherOrder  + ' />';
		html += '<div style="display:inline; width:50%;"> Control order: </div> <input id="controlOrder" type="text"  style="margin-top: 1.5rem; width:50%; letter-spacing: .2em;" value=' + controlOrder + ' />';
		html += '<div style="display:inline; width:50%;"> Index order:   </div> <input id="indexOrder"   type="text"  style="margin-top: 1.5rem; width:50%; letter-spacing: .2em;" value=' + indexOrder   + ' />';

		html += '<div style="text-align:center;"><button id="dialogSubmit" class="clrButton" style="text-align:center; margin-top:5px; margin-bottom:10px; margin-left:auto; margin-right:auto; width:40%">Set Rotor Order</button></div>';
		return html;
} // end of changeRotorOrderHTML

/* manageStoredRotorSetsHTML
 * Generate HTML for the manage stored rotor sets dialog.
 *
 */
function manageStoredRotorSetsHTML() {

	let rotorSets = storageManager.listRotorSets();
	
	// Add the default rotor set to the top of the list.

	rotorSets.unshift(Rotor.DEFAULT_ROTOR_SET_NAME);

	if(!rotorSets) {return [false, "No rotor sets found"]; }

	let html  = "<h2 style='margin-bottom: auto; text-align:center; clear:both;'>Manage Rotor Sets</h2>";
		html += "<div style='width:80%; border: 1px solid black; overflow-y:scroll; height: 300px; margin: auto;'>";
		html += "<table id='storedRotorSets'>";
		html += "<tr><th>Set Name</th><th>Actions</th></tr>";
				
	for(const name of rotorSets) {
		let itemInstall;
		
		if(Rotor.getActiveRotorSetName() == name) {
			itemInstall    = '<span class="green">Active</span>';
		} else {
			itemInstall = "<input id='" + name + ":install'  title='install'  class='rotorSetActions' type='image' src='img/installSetIcon.svg'  value='install'  alt='install'>";			
		}
		
		let itemDownload   = "<input id='" + name + ":download' title='download' class='rotorSetActions' type='image' src='img/downloadSetIcon.svg' value='download' alt='download'>";
		let itemDelete     = "<input id='" + name + ":delete'   title='delete'   class='rotorSetActions' type='image' src='img/deleteSetIcon.svg'   value='delete'   alt='delete'>";
//		let itemInstall    = "<input id='" + name + ":install'  title='install'  class='rotorSetActions' type='image' src='img/installSetIcon.svg'  value='install'  alt='install'>";

		html += "<tr><td>" + name + "</td><td>" + itemInstall + itemDownload + itemDelete + "</td></tr>";
	}
	html += "</table>";
	html += "</div>";
	return [true, html];
} // end of manageStoredRotorSetsHTML

/** manageRotorSet
 *  Called by the rotorSetActions listener to execute the various rotor set management actions:
 *  install, download, and delete.
 *  Also called by the rotorOrder action listener to execute changing rotor orderings.
 * 
 *  @author	Stephen R. MacMinn, April 2024
 * 
 *  @return none	Posts a message to the dialog messageWindow.
 * 
 *  @param	string	name	Name of the rotor set to act on.
 *  @param	string	func	Action to take. (download, install, delete, order)
 * 
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 12 Apr 2024		1.00	SRM		First release
 *
 */
function manageRotorSet(name, func) {
	let html;
	
	// Excecute the desired function.
	// Each case in the switch executes the function and then
	// rewrites the proper dialog for the user.
	
	switch(func) {
		
		/* download */
		
		case 'download': 
		{
			let rotorSet = [];
			if(name == Rotor.DEFAULT_ROTOR_SET_NAME) {
				rotorSet = [...Rotor.defaultWIRING];
			} else {
				rotorSet = storageManager.getRotorSet(name);
			}
	
			let rotorSetString = csvStringify(rotorSet);
			
			// Create element with <a> tag
			let link = document.createElement("a");

			// Create a blog object with the file content which you want to add to the file
			const file = new Blob([rotorSetString], { type: 'text/plain' });
			
			// Add file content in the object URL
			link.href = URL.createObjectURL(file);
			
			// Add file name
			link.download = name + ".csv";
			
			// Add click event to <a> tag to save file.
			link.click();
			URL.revokeObjectURL(link.href);
			postMessage([true, "File '" + name + ".csv' downloaded"]);
			document.getElementById("dialogContent").innerHTML =  manageStoredRotorSetsHTML();
			bindManageListeners();
			break;
		}
			
		/* delete */
		
		case 'delete':
		{
			if(name == Rotor._activeRotorSetName) {
				postMessage([false, "Unable to delete the active rotor set."]);
				return;
			}
			let result = storageManager.deleteRotorSet(name);
			postMessage(result);
			document.getElementById("dialogContent").innerHTML =  manageStoredRotorSetsHTML();
			bindManageListeners();
			break;
		}
		
		/* install */
		
		case 'install':
		{	
			let rotorSet= storageManager.getRotorSet(name);
			result = Rotor.installRotorSet(name, rotorSet);
			postMessage(result);
			document.getElementById("dialogContent").innerHTML =  manageStoredRotorSetsHTML();
			bindManageListeners();
			break;
		}
		
		/* order */
		
		case 'order':
		{
			let newCipherOrder  = document.getElementById('cipherOrder').value.toUpperCase();
			let newControlOrder = document.getElementById('controlOrder').value.toUpperCase();
			let newIndexOrder   = document.getElementById('indexOrder').value.toUpperCase();
			
			// If any of the fields are empty, use their default values.
			
			if(!newCipherOrder ) { newCipherOrder  = defaultCipherOrder;  }
			if(!newControlOrder) { newControlOrder = defaultControlOrder; }
			if(!newIndexOrder  ) { newIndexOrder   = defaultIndexOrder;   }
			
			result = validateRotorOrder(newCipherOrder);
			if(!result[0]) {
				postMessage(result);
				return;
			} else {
				cipherOrder = newCipherOrder;
			}
			
			result = validateRotorOrder(newControlOrder);
			if(!result[0]) {
				postMessage(result);
				return;
			} else {
				controlOrder = newControlOrder;
			}
			
			result = validateRotorOrder(newIndexOrder);
			if(!result[0]) {
				postMessage(result);
				return;
			} else {
				indexOrder = newIndexOrder;
			}
			
			// All new orderings are valid.  Install them.
			
			cage = new RotorCage(cipherOrder, controlOrder, indexOrder);
			cage.zeroize();
	    	cage.setIndexBankPos("00000");		
			postMessage([true, 'New rotor ordering installed.']);
			document.getElementById("dialogContent").innerHTML =  changeRotorOrderHTML();
			break;
		}
			
		default:
			postMessage([false, "Unrecognized action: '" + name + "'"]);
	}
	return;	
}  // end of manageRotorSet

/** validateRotorOrder
 *  A valid rotor ordering is:
 * 		10 characters 
 *		Alternates digits with "N" or "R" e.g., '0N1N2R3N4R'
 *
 * @param	string	order	Rotor ordering string to validate.
 * @return	array			[true|false, message string]
 * 
 * Author:	Stephen R. MacMinn
 * 
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 12 Apr 2024		1.00	SRM		First release
 * 
 */
function validateRotorOrder(order) {
	
	if(order.length != 10) 			 	 { return [false, 'Rotor order string must be 10 characters, ' + order + ' is not.']; }
	if(!(/([0-9](R|N)){5}/.test(order))) { return [false, 'Improper rotor order string'];}

	return [true, 'Rotor order string validated.'];
}

/** validateRotorSet
 *  A valid rotor set must:
 * 		- contain 10 rotors.
 * 		- Each rotor must contain 26 elemennts.
 * 		- Each element must be a letter.
 * 		- All the elements in a rotor must be distinct.
 * 
 *  @param	array	rotorSet	A rotor set array.
 *  @return	array				[true|false, message]
 * 
 * Author:	Stephen R. MacMinn
 * 
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 12 Apr 2024		1.00	SRM		First release
 *
 */

function validateRotorSet(rotorSet) {
	
	if(rotorSet.length != 10) {
		return [false, 'Set must contain 10 rotors'];
	}
	
	// Valid rotor elements are all letters, but not ' '.
	
	let validRotorElementsSet = letterKeySet;
	
	let i;
	for(i=0; i<10; i++) {
		let rotor = rotorSet[i];
		if(rotor.length != 26) {
			return [false, "Upload Canceled: A rotor must have 26 elements.  Rotor " + i + " does not."];
		}

		const checkSet = new Set(rotor);		// checkSet will only take distinct elements from rotor.
		if(checkSet.size != 26) {
			return [false, 'Upload Canceled: Elements of a rotor must be distinct.  Rotor ' + i + ' has duplicate elements']
		}
		
		for(j=0; j<26; j++) {
			rotor[j] = rotor[j].toUpperCase();
			if(!(validRotorElementsSet.has(rotor[j]))) {
				return [false, 'Upload Canceled: Elements of a rotor must all be letters. Element ' + j + ' of rotor ' + i + ' is invalid.']
			}
		}
	}
	return [true, "Rotor set is valid"]
}

/** bindManageListeners
 *  Binds listeners to manage rotors dialog rotorSetActions elements: 
 *  install, download, and delete, for each rotor set.
 * 
 *  This is done in a function because it needs to be called from two different places: 
 *  When we first set up the rotor management dialog, and when we rewrite it after a change.
 * 
 * @author: Stephen R. MacMinn, April 2024
 * 
 * @param	none
 * @return	none	Event listeners bound to all 'rotorSetActions' DOM elements
 * 
 * Author:	Stephen R. MacMinn
 * 
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 12 Apr 2024		1.00	SRM		First release
 *
 */
function bindManageListeners() {

	// Fails if there are no stored rotor sets.
	
	let result = manageStoredRotorSetsHTML();
	if(!result[0]) {
		postMessage(result);
		return;
	}
	
	// Display the dialog.
	
	document.getElementById("dialogContent").innerHTML = result[1];
	
	/* 
	 * Add the event listeners.
	 * When the event fires, get the id, which will be of the form "name:action"
	 * Split this to get the name.  Then pass the name and function to manageRotorSet()
	 */
	
	let actions = document.getElementsByClassName('rotorSetActions');
	for(let action of actions) {
		action.addEventListener('click', function (e) {
			let name = e.currentTarget.id;
			let b = name.split(":");
			name = b[0];
			let func = e.currentTarget.value;
			manageRotorSet(name, func);
		});
	} 
}

/** enableKeySound - Signal to turn on/off keyboard clicks.
 *  
 *  @param		b		boolean		True to enable / false to disable.
 *  @return		none
 * 
 *  @author	Stephen R. MacMinn
 * 
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 12 Apr 2024		1.00	SRM		First release
 *
 */ 
function enableKeySound(b) {
	keySoundOn = b;
}

/***********************************************************************
 * action() handles component events. 
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 * 
 * Javascript version, Completely rewritten to use HTML DOM and javascript eventListeners.
 * Copyright (C) 2024, by Stephen R. MacMinn
 *
 * SUMMARY:
 * action() defines the listeners for all component events.  Since all activity of the machine is driven
 * by the switches this method emulates all of the wiring that is not in RotorCage.  Most of
 * the detail here is less interesting than looking at ECMCycle and the RotorCage.
 *
 * The switch action table:
 *
 *          P                       R                       E
 *  Zeroize Blank key only          Blank key only          Blank key only
 *          Counter updates         Counter updates         Counter updates
 *          No Rotor Moves          Zeroize rotors          Zeroize rotors
 *          Prints Blank            No printing             Prints Blanks
 *
 *  Operate All printable keys      1-5 and Blank keys      26 alpha keys
 *          Counter updates         Counter updates         Counter updates
 *          No Rotor Moves          1-5 move cipher&control Control & cipher move
 *          Blank is a space?       Blank moves cipher only X, Z, space & Blank handling
 *          All printable keys      No printing             Printing
 *                                                          
 *  When switching from E to R, or P to R, one to four spaces are output.
 *  Cipher text is printed in 5 character groups separated by spaces.
 *  
 *  Decipher is same as Encipher except Z, X and Space Bar and direction.
 *  During decipher, the entered text is written to the cipher text tape, and
 *  the deciphered text is writeen to the plaintext tape.
 * 
 *  During encipher, Z is converted to X, and Space Bar is converted to Z, and
 *          rotors are read from left to right.
 *      During decipher, Z is converted to space, Space Bar is ignored, and
 *          rotors are read from right to left.
 *  During both Decipher and Encipher, Blank moves rotors, prints a space.
 *
 *  If 889/2900 is in the middle, no keys work.
 *  If 889/2900 knob and switch do not match, no keys work
 *  If Zeroize/Operate switch is in the middle, no keys work
 *
 *  cipherCount implements a counter that increments on each encipher or deciphered 
 *      character.  It is cleared whenever the 1st or 5th rotor turns or the master switch
 *      is not E or D.  If it reaches 21, the machine is locked up until the master switch is
 *      rotated out of E. This was added postwar, probably to detect improperly positioned
 *      index rotors.  If the index rotors are not positioned properly there will be no cipher
 *      rotor movement.  In the program this is possible in a practicle sense only if straight
 *      through test control rotors are used.
 *
 *  Also see the curatorial report, SIGKKK and the Army history of converter.
 * 
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96 	 1.00   	RSP 	First release.
 * 	 Apr 2024	 1.1		SRM		Completely rewritten in javascript using
 * 									event listeners and the browser DOM.
 *      
 ***********************************************************************/

 function action() {
	
	// 
	// Options handlers 
	//

	/* Open the about dialog	*/
	document.getElementById("aboutDialogOpener").addEventListener('click', function(e) {
		document.getElementById("overlay").style.visibility = "visible";
		document.getElementById("dialogContent").innerHTML = aboutDialog;
		document.getElementById("lightbox").style.visibility = "visible";
		enableKeySound(false);
		simulatorMode = MODE_DIALOG;
	});
	
	/* Open rotor upload dialog */
	
	document.getElementById("uploadDialogOpener").addEventListener('click', function(e) {
		document.getElementById("overlay").style.visibility = "visible";
		document.getElementById("dialogContent").innerHTML = fileUploadDialog;
		document.getElementById("lightbox").style.visibility = "visible";
		enableKeySound(false);
		simulatorMode = MODE_DIALOG;
		
		// Once the dialog is in the DOM, add the "submit" event listener.
		document.getElementById("dialogSubmit").addEventListener('click', function(e) {
			let file = document.getElementById("uploadRotorsFile").files[0];
			let rotorSetName = document.getElementById("rotorSetName").value;
			processRotorSetUpload(rotorSetName, file);
		});
	});
	
	
	/* Open the rotor order dialog */
	
	document.getElementById("rotorOrderDialogOpener").addEventListener('click', function(e) {
		document.getElementById("overlay").style.visibility = "visible";
		document.getElementById("dialogContent").innerHTML = changeRotorOrderHTML();
		document.getElementById("lightbox").style.visibility = "visible";
		enableKeySound(false);
		simulatorMode = MODE_DIALOG;
		
		// Once the dialog is in the DOM, add the "submit" event listener.
		document.getElementById("dialogSubmit").addEventListener('click', function(e) {
			manageRotorSet('', 'order');
		});

	});

	/* Open the manage rotor sets dialog */
				
	document.getElementById("manageRotorSetsDialogOpener").addEventListener('click', function(e) {
		document.getElementById("overlay").style.visibility = "visible";
		document.getElementById("dialogContent").innerHTML = fileUploadDialog;
		document.getElementById("lightbox").style.visibility = "visible";
		simulatorMode = MODE_DIALOG;

		// Fails if there are no stored rotor sets.
		
		let result = manageStoredRotorSetsHTML();
		if(!result[0]) {
			postMessage(result);
			return;
		}
		
		// Display the dialog.
		
		document.getElementById("dialogContent").innerHTML = result[1];
		bindManageListeners();
		enableKeySound(false);	
	});
	

	/* Non-Keyboard Button handlers */

	// Clear counter button - this is mechanical and works even if the machine is off.
	document.getElementById('clearCounter').addEventListener('click', function(e) {
		count = 0;
		document.getElementById(countDispId).value = '';
	});
	
	/* Clear tapes button */ 
	// this does not exist as a button on the machine
	// but is added as a convenience in the simulator.
	
	document.getElementById(tearTapeId).addEventListener('click', function(e) {
	    document.getElementById(paperTapeId).innerText = "";
    });

	document.getElementById(clearKeyboardDisplayId).addEventListener('click', function(e) {
	    document.getElementById(keyboardDisplayId).innerText = "";
    });
	
	/* Copy paper tape and plaintext tape */
	
	document.getElementById(copyPaperTapeId).addEventListener('click', function(e) {
		copyText(paperTapeId);
		alert("Paper Tape copied");
	});
	
	document.getElementById(copykeyboardDisplayId).addEventListener('click', function(e) {
		copyText(keyboardDisplayId);
		alert("Keyboard Display copied");
	});

 	/* Process letter keys. */
 	
	let keys = document.getElementsByClassName(keyClass);

	for(let key of keys) {
		key.addEventListener('mousedown', function (e) {
			
			// Catch only left clicks
			
			if (e.button != 0) { return; }
			
			e.preventDefault();
			
			if(keySoundOn) {
				clickAudio.play();
			}

			// Shrink the key for visual effect.
			
			let elem = document.getElementById(e.currentTarget.id);
			let key = elem.name;			
			elem.className = smallKeyClass;	// Shrink the key

			letter_action(key);			
		});		

		// When mouse is released, restore the larger key.
		
		key.addEventListener('mouseup',  function (e) {
				document.getElementById(e.currentTarget.id).className = keyClass;		// Make the key big again.
		});
	}

	/* Handle physical keyboard events */
		
	document.addEventListener('keydown', function(e) {
		if(simulatorMode != MODE_RUN) { return; }
		
		// Changes to the rotor position fields are handled separately.
		
		let elemId = document.activeElement.id;
		if(elemId == cipherDispId  ||
		   elemId == controlDispId ||
		   elemId == indexDispId)  { 
			return; 
		}

		let key = e.key.toUpperCase();
		
		// Only letters and space.
		
		if(letterKeySet.has(key) || key == ' ') {
			if(keySoundOn) {
				clickAudio.play();
			}
			letter_action(key);			
		}
	});	
	
   // 
   // Handle choices, i.e. the switches. 
   //

	/* masterSwitch */
	       
    document.getElementById(masterSwitchId).addEventListener('change', function(e) {
		ms = document.getElementById(masterSwitchId).value;
		if ("OPR".indexOf(ms) >= 0) {
            cage.cipherCount = 0;   // This is used to detect lack of cipher movement.
        }

        if (ms === "O") {   		// Turn the machine off.                    
            oldms = "O";    		// oldms is used to detect switch transitions used below
            encPaperCount = 0;  	// This is used to put cipher text in 5 char groups.
            return true;
        }
            
        // Zero the encrypt paper count when not in encrypt.
        // When you change from P to R, or E to R, between 1 and 4 
        // spaces are output on the real machine while resetting the cam
        // that spaces 5 character groups during encrypt. To simulate this
        // behavior this program looks at count, this is not quite the same as
        // keeping track of the real state of the cam, but close enough. This
        // spacing is an artifact, not an essential element of the machine.

        if ((ms === "R") && 
            (oldms === "P") || (oldms === "E")) {
            encPaperCount = count % 5;
            while ((encPaperCount++ % 5) != 0) {
                appendToTape(paperTapeId, " ");
            }
            encPaperCount = 0;
            oldms = ms;
            return true;
        }
                                
        oldms = ms; // oldms is used to detect switch transitions above.
	});	// end of masterSwitch
	
            
    /* Handle Machine Type switch. */

    // Some future version of the program might have both the knob and the switch.  
    // When this happens, both knob and switch must agree for anything to happen.
    
    document.getElementById(machSwitchId).addEventListener('change', function(e) {
 		let s = document.getElementById(machSwitchId).value;
   
        if (s === "CSP 889") {
            machType = CSP889;
        
        } else if (s === "CSP 2900") {
            machType = CSP2900;

        } else {
            machType = CSPNONE;

        }
        return true;
    });   // end of machSwitch
            
    /* Handle Zeroize switch. */
          
    document.getElementById(zeroSwitchId).addEventListener('change', function(e) {
 		zs = document.getElementById(zeroSwitchId).value;
        return true;
    });   // end of zeroSwitch 
     
    //
    //  Handle TextFields 
    //

    // Cipher rotor display.                       
 
    document.getElementById(cipherDispId).addEventListener('change', function(e) {
		let x = e.currentTarget.value.toUpperCase();
		processTextField(cipherDispId, x);
	});

	// Control rotor display
	
    document.getElementById(controlDispId).addEventListener('change', function(e) {
		let x = e.currentTarget.value.toUpperCase();
		processTextField(controlDispId, x);
	});

	// Index rotor display
	
    document.getElementById(indexDispId).addEventListener('change', function(e) {
		let x = e.currentTarget.value.toUpperCase();
		processTextField(indexDispId, x);
	});        
} // end of actions


/** copyText
 *  Copies text from the selected element to the clipboard
 * 
 * @author	Stephen R. MacMinn, April 2024
 * 
 * @param	string	id		DOM id of the element to copy.
 * 
 * @return	none			Any text in the element is on the clipboard
 *
 */
function copyText(id) {
	var elem = document.getElementById(id);
	navigator.clipboard.writeText(elem.innerText);
}

/** processTextField
 *  Called by the input field event listeners for the rotor position fields.
 *  Updates the rotor position to the selected value.
 * 
 *  @param	string	id		DOM id of the changed field.
 *  @param	string	value	New value
 * 
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    	RSP 	First release.
 * 12 Apr 2024	1.1			SRM		Translated to javascript.
 * 
 */
function processTextField(id, value) {
    if ((id === cipherDispId) || (id === controlDispId)) {
    
        // Make sure the string is 5 characters long, i.e. 5 rotors. Fill
        // with letter "O" if too short, truncate if too long.              
        for (; value.length < 5; value = value + "O") {}
        
        if (value.length > 5) {
            value = value.substring(0, 5);
        }
            
        // check to makes sure each of the positions is alpha.              
        value = value.toUpperCase();

        for (i = 0; i < 5; i++) {
            if (('A' > value.charAt(i)) ||
            	('Z' < value.charAt(i)) ) {
            
                // Replace non-alpha positions with letter "O".
                value = value.replace(value.charAt(i), 'O');  
            }
        }

        
        if (id === cipherDispId) {
            cage.setCipherBankPos(value);   // Update the cipher rotor positions.
            setValue(cipherDispId, value);   // Update the display

		// id == controlDisp
        } else {  
            cage.setControlBankPos(value);  // Update the control rotor positions.
            setValue(controlDispId, value);  // Update the display
        }
        return true;
    }   // end of Cipher and Control Display handling

    // Index rotor display.         
    if (id === indexDispId) {
    
        // Make sure the string is 5 characters long, i.e. 5 rotors. Fill
        // with zero if too short, truncate if too long.
        for (; value.length < 5; value = value + "0") {
            }
        if (value.length > 5) {
            value = value.substring(0, 5);
            }
            
        // check to makes sure each of the positions is between 0 and 9.
        for (i = 0; i < 5; i++) {
            if ("0123456789".indexOf(value.charAt(i)) < 0) {    // Not 0-9?
                // Replace non-numeric positions with zero.
                value = value.replace(value.charAt(i), '0');  
                }
            }
        cage.setIndexBankPos(value);    // Update the index rotor positions.
        setValue(indexDispId, value);       // Update the display
        return true;
    }   // end of indexDisp handling           

    return false;   // we did not handle this TextField
}   // End of text field handler


/** letter_action
 *  Process events from either the virtual or real keyboard.
 * 
 * @param	string	key		A string representing the entered key.  The spacebar is
 * 							represented by "SPACE" and the blank key by "BLANK"
 * 
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    	RSP 	First release.
 * 12 Apr 2024	1.1			SRM		Translated to javascript.
 *
 */
function letter_action(key) {
	
    // If the master switch (ms) is off OR the machine switch is in the middle,
	// or zeroize switch (zs) is in the middle, do nothing by returning.

	 ms = document.getElementById(masterSwitchId).value;
	 zs = document.getElementById(zeroSwitchId).value;

	if ((ms === "O") || (machType == CSPNONE) || (zs === "")) {
		return;
	}

	key = key.toUpperCase();
	
	if(key === "SPACE") { key = ' ';}

  
	// Handle the cipherCounter. If the cipher counter is 21, return with no action.
	// (see the "action" function description for details)
	
	if (cage.cipherCount >= 21) {
	    return true;
	}
	                                        
                
	/* Handle any non-repeat keyboard button */
	
	// First handle Zeroize switch in R, P, E and D positions of master switch.
	if (zs === "Zeroize") {     // When Zeroize is selected,
		buttonIfZeroize(key);
	} else {
		buttonIfNotZeroize(key);
	}
}

/** buttonIfZeroize
 *  Process letter keys if "zeroize" is set.
 * 
 *  @param	string	key		The key value.
 *
 *  
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    	RSP 	First release.
 * 12 Apr 2024	1.1			SRM		Translated to javascript.
 * 
 */

function buttonIfZeroize(key) {
			
	// only respond to the Blank key, and
	// only move the rotors if Reset, Encrypt or Decrypt. Note that you
	// should really only use Zeroize with Reset, the rest of this behavior
	// is a quirk.

	if(key !== BLANK_KEY) { return; }	// not a Blank key in zeroize, then do nothing.
	if ("RED".indexOf(ms) >= 0) {
                    
        // The Blank key will advance each of the rotors position until
        // they are all on the "O" position.
        for (let i = 0; i < 5; i++) {   // There are 5 rotors in a bank.
            
            // CSP 889 cipher rotors all move in the normal clockwise rotation
            if (machType == CSP889) {   
                if (cage.cipherBank[i].pos != 'O'.charCodeAt(0) - 'A'.charCodeAt(0)) {
                    cage.cipherBank[i].rotCW();
                    // update the cipher bank display.
                    setValue(cipherDispId, cage.cipherBankPosToString());

                    // clear the cipher rotor movement counter if the first or
                    // last rotor turn.
                    if (i == 0 || i == 4) {
                        cage.cipherCount = 0;
                    }
                }
            }
            
            // CSP 2900's 2nd and 4th cipher rotors turn counter clockwise
            if (machType == CSP2900) {  
                if (cage.cipherBank[i].pos != 'O'.charCodeAt(0) - 'A'.charCodeAt(0)) {
                    if ( (i == 0) || (i == 2) || (i ==4) )
                        cage.cipherBank[i].rotCW();
                    else 
                        cage.cipherBank[i].rotCCW();
                    // update the cipher bank display.
                    setValue(cipherDispId, cage.cipherBankPosToString());

                    // clear the cipher rotor movement counter if the first or
                    // last rotor turn.
                    if (i == 0 || i == 4) {
                        cage.cipherCount = 0;
                    }
                }
            }
            
        	// Both machines rotate the control bank in the same direction
            if (cage.controlBank[i].pos != 'O'.charCodeAt(0) - 'A'.charCodeAt(0)) {
                cage.controlBank[i].rotCW();
                
                // update the control bank display.
                setValue(controlDispId, cage.controlBankPosToString());
            }
        }   // end of the 5 rotor for loop
            
        // Increment the counter and display it.
        document.getElementById(countDispId).value = ++count;
                
        // Another quirk, not only does Blank work in zeroize in E and D,
        // but it also prints a space.
                              
        if ("ED".indexOf(ms) >= 0){
			
            // Also add a space to the paper tape.

            appendToTape(paperTapeId, " ");
            
        }
        return true;
   }   // end of a zeroize with the Blank key in R, E or D
                    
    // The doc indicates that with zeroize and plaintext the machine should 
    // do nothing, but on the real machine a Blank key spaces the paper tape
    // and updates the counter.
    
    if (ms === "P"){
        // Increment the counter and display it.
        setValue(countDispId, ++count)
        appendToTape(paperTapeId, " ");
        return true;
    }   // end of zeroize and plaintext with Blank key.
}   // end of zeroize switch


/** buttonIfNotZeroize
 *  Process letter keys if "zeroize" NOT set.  In this case,
 *  The machine is in Operate (not Zeroize) mode,
 *  This function handles R, P, E and D master switch positions.
 * 
 *  @param	string	key		The key value.
 *
 *  
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    	RSP 	First release.
 * 12 Apr 2024	1.1			SRM		Translated to javascript.
 * 
 */

function buttonIfNotZeroize(key) { 
	           
    let j;
    
    // In the Reset position of master switch, pressing number keys 1 through 5
    // advances the corresponding control and cipher rotors. It is used to set the
    // rotor position to a key.  Another quirk is that the Blank key moves the cipher
    // rotors.
    
    if (ms === "R") {
		
        // The 1-5 keys work in R position
        j = "12345".indexOf(key);
        if ( j >= 0) {          // indexOf is positive if key is found
        
            // Rotate 1 to 4 cipher rotors.  
            cage.cipherBankUpdate(machType);
            // update the cipher bank display.  
            setValue(cipherDispId, cage.cipherBankPosToString())

            // change the position of a control rotor
            cage.controlBank[j].rotCW();
            // update its display   
            setValue(controlDispId, cage.controlBankPosToString())

            // Increment the counter and display it.
            setValue(countDispId, ++count)
            return;
        }
            
        // The doc indicates that nothing should happen here, but Blank moves the
        // cipher rotors, this is a quirk, not a feature of the real hardware.

		if(key === BLANK_KEY) {
			 
            // Rotate 1 to 4 cipher rotors. 
            cage.cipherBankUpdate(machType);

            // update the cipher bank display.  
         	setValue(cipherDispId, cage.cipherBankPosToString());
            
            // Increment the counter and display it.
            setValue(countDispId, ++count);

            return;
        }
        // Keys other than 1-5 and Blank do nothing
        return;
    }
                  
                              
    // Plaintext switch position
    if (ms === "P") {
        // Convert Space Bar to a single space.
        if ((key === BLANK_KEY) || (key === " ")) {
            key = " ";
        }
        setValue(countDispId, ++count);

		appendToTape(paperTapeId, key);
		appendToTape(keyboardDisplayId, key)
        return;
    }
        
    /* Now encipher or decipher.  Only character keys, number keys, dash,
    and Blank and Space Bar are left. */
    
    // Number keys and the dash key are ignored during encipher or decipher.
    if ("1234567890-".indexOf(key) >= 0) {
        return;    // ignore the key.
    }
        
    // Blank in E or D has the odd behavior of updating the cipher and control rotor
    // banks and printing a space.
    if (key === BLANK_KEY) {
		
        // ECM cycle will update both cipher and control rotor banks
        let enkey = ECMcycle(" ", ENCRYPT, machType); 
        setValue(countDispId, ++count);

		appendToTape(paperTapeId, enkey);
		appendToTape(keyboardDisplayId, ' ');
        encPaperCount++;
        return true;
    }
                
    /* encipher */
    
    if (ms === "E") {
        // Convert Z to X. There are only 26 cipher text characters.
        
        if (key === "Z") { key = "X"; }
        
        // Convert Space Bar to Z. Spaces are more important than Z. Note that the
        // deciphered plaintext can never have a Z.
        
        if ( key === ' ') { key = "Z"; }
        let enKey = ECMcycle(key, ENCRYPT, machType);
 		setValue(countDispId, ++count);

        // Add a space to the paper tape if needed to generate 5 character groups. 
        // Do not add space the first time through, i.e. encPaperCount = 0.
        
        if ((encPaperCount != 0) && ((encPaperCount % 5) == 0)) {   
			appendToTape(paperTapeId, " " + enKey); 	// Print the space and ciphertext.
		} else {
            appendToTape(paperTapeId, enKey);			// Print the cipher text.
        }
        appendToTape(keyboardDisplayId, key);				// Print the plaintext
        
        encPaperCount++;
        cage.cipherCount++;
        return true;
    }   // end of encipher
        
    /* decipher */
    
    if (ms === "D") {
        if (key === " ") { // Ignore Space Bar
            return true;
            }
        let enKey = ECMcycle(key, DECRYPT, machType);

        // Convert Z to Space.
        if (enKey === "Z") {
            enKey = " ";
        }
        
        // Reverse the display
        // "paperTape" is now the typed-in (enciphered) text."
        // "plainTextTape is the deciphered text.
        
        setValue(countDispId, ++count);
		appendToTape(keyboardDisplayId, key);
		appendToTape(paperTapeId, enKey);

        cage.cipherCount++;
        return true;
    }   // end of decipher
 }   // end of buttonIfNotZeroize.

/** setText
 *  Populate a innerText of an HTML element with text
 *
 *  @author	Stephen R. MacMinn
 * 
 *  @param	string	id			DOM id of the field to populate.
 *  @param	string	text		Text to populate.
 * 
 */        
function setText(id, property, text) {
	document.getElementById(id).innerText = text;
}

/** setValue
 *  Populate a content of an HTML input element with text
 *
 *  @author	Stephen R. MacMinn
 * 
 *  @param	string	id			DOM id of the field to populate.
 *  @param	string	text		Text to populate.
 * 
 */
function setValue(id, text) {
	document.getElementById(id).value = text;
}

/** appendToTape 
 *  Appends text to one of the two tape displays.
 * 
 *  @author	Stephen R. MacMinn
 * 
 *  @param	string	id		DOM id of the field to populate.
 *  @param	string	text	Text to populate.
 */
function appendToTape(id, text) {
	let tapeText = document.getElementById(id).innerHTML + text;
	document.getElementById(id).innerHTML = tapeText;
}

/***********************************************************************
 * ECMcycle enciphers or deciphers a single character.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 * 
 * Javascript code copyright (C) 2024, by Stephen R. MacMinn
 *
 * SUMMARY:
 * Enciphers or deciphers a single character through the RotorCage object.
 *
 * Control and cipher rotor bank contacts are labeled as if a rotor were positioned
 * with letter "A" on top, not reversed, as printed (i.e. increasing clockwise.) 
 * Internally these positions are represented by ints from 0-25. The pos of a rotor 
 * is its displacement from A on top.  For later reference, index rotor contacts are
 * similarly labeled to match an index rotor with zero on top (10,20,30,40,50 on top),
 * not reversed, as printed (i.e. increasing counter clockwise.)
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96 	    1.00    RSP 	First release.
 * 14 Apr 2024		1.10	SRM 	translated to javascript.
 *      
 ***********************************************************************/
 function ECMcycle(s, direction, machine) {
    let inx, out;
    let sout;
            
    // convert from keyboard string to internal integer representation.
    
    inx = "ABCDEFGHIJKLMNOPQRSTUVWXYZ".indexOf(s);   			// keyboard/printer contact
        // 01234567890123456789012345                			// internal representation result
    
    out = cage.cipherBankPath(direction,inx);        			// encipher or decipher the character
    
    // Convert from internal integer representation to string.  This is the inverse of
    // the lookup done before the encipher or decipher.  Note that substring goes from
    // the first parameter to the second parameter - 1.
    
    sout = "ABCDEFGHIJKLMNOPQRSTUVWXYZ".substring(out,out+1);
        //  01234567890123456789012345          				// internal representation result

    cage.cipherBankUpdate(machine);         					// Rotate 1 to 4 cipher rotors. 
	setValue(cipherDispId, cage.cipherBankPosToString())

    cage.controlBankUpdate();               					// Rotate the control rotors in meter fashion.
	setValue(controlDispId, cage.controlBankPosToString()) 
 
    return(sout);
}   //end of ECMcycle
