/***********************************************************************
 * Rotor object emulates an ECM rotor.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * REVISION HISTORY:
 *
 *   Date   Version 	By  	Purpose of Revision
 * -------- ------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
class Rotor {
    static ENCRYPT = false; DECRYPT = true;
    static RIGHT = 1;  // Rotor wiring table from right to left
    static LEFT  = 0;  // Rotor wiring table from left to right
    
    // This table represents the physical rotor wiring for large rotors.
    // Each machine was supplied with 10 large, 26 contact rotors that could be used as
    // either control or cipher rotors.  I do not have any authentic rotor wirings, these 
    // were created created with a shift register type generator. This table should be 10
    // items long. i.e. the two test rotors are normally commented out.  The table's index
    // is the left side of the rotor, the entry in the table the right. 
    // I.e. right side = WIRING[][left side]. Real rotors are labeled on the outside
    // circumference increasing in the clockwise direction.

    static WIRING = [
        ['A','B','C','D','E','F','G','H','I','J','K','L','M',   	//[0][] Straight through
        'N','O','P','Q','R','S','T','U','V','W','X','Y','Z'],   	//      test rotor.
//d2        {'B','C','D','E','F','G','H','I','J','K','L','M','N',   //[0][] Shift by one
//d2        'O','P','Q','R','S','T','U','V','W','X','Y','Z','A'},   //      test rotor.
// Simulated random rotors are below:
//d1        ['Y','C','H','L','Q','S','U','G','B','D','I','X','N',   //[0][] 
//d1        'Z','K','E','R','P','V','J','T','A','W','F','O','M'],
        ['I','N','P','X','B','W','E','T','G','U','Y','S','A',   	//[1][] Rotor wired Oct 96 in
        'O','C','H','V','L','D','M','Q','K','Z','J','F','R'],   	// 		the real machine.
        ['W','N','D','R','I','O','Z','P','T','A','X','H','F',   	//[2][] Rotor wired Sep 97 in
        'J','Y','Q','B','M','S','V','E','K','U','C','G','L'],   	// 		the real machine.
        ['T','Z','G','H','O','B','K','R','V','U','X','L','Q',   	//[3][]
        'D','M','P','N','F','W','C','J','Y','E','I','A','S'],
        ['Y','W','T','A','H','R','Q','J','V','L','C','E','X',   	//[4][]
        'U','N','G','B','I','P','Z','M','S','D','F','O','K'],
        ['Q','S','L','R','B','T','E','K','O','G','A','I','C',   	//[5][]
        'F','W','Y','V','M','H','J','N','X','Z','U','D','P'],
        ['C','H','J','D','Q','I','G','N','B','S','A','K','V',   	//[6][]
        'T','U','O','X','F','W','L','E','P','R','M','Z','Y'],
        ['C','D','F','A','J','X','T','I','M','N','B','E','Q',   	//[7][]
        'H','S','U','G','R','Y','L','W','Z','K','V','P','O'],
        ['X','H','F','E','S','Z','D','N','R','B','C','G','K',   	//[8][]
        'Q','I','J','L','T','V','M','U','O','Y','A','P','W'],
        ['E','Z','J','Q','X','M','O','G','Y','T','C','S','F',   	//[9][] 
        'R','I','U','P','V','N','A','D','L','H','W','B','K']
        ];        

        //  Just for quick reference.
    	//  ABCDEFGHIJKLMNOPQRSTUVWXYZ  // letter
        //  01234567890123456789012345  // internal representation result
    
    static defaultWIRING = [...Rotor.WIRING];
	static DEFAULT_ROTOR_SET_NAME = "Default";
    static _activeRotorSetName = Rotor.DEFAULT_ROTOR_SET_NAME;
  
    // This table represents the physical rotor wiring for small, index rotors.
    // This table should be 5 items long.  I.e. the two test rotors are normally
    // commented out.  These are real rotor wirings as found in the CSP-889/2900
    // now aboard USS Pampanito.  We do not know when they served.  They are all
    // embossed with "CTT 68AAC" on the edge.  Note that index rotors labeling
    // on their circumference increases in the counter clockwise direction.
	
    static INDEXWIRING = [
    //  0 1 2 3 4 5 6 7 8 9 Left side of rotor, below is right side.
//d3    [0,1,2,3,4,5,6,7,8,9],  // [0][] Straight through test rotor.
//d3    [1,2,3,4,5,6,7,8,9,0],  // [0][] A shift by one test rotor
    // The real rotors are below:
        [7,5,9,1,4,8,2,6,3,0],  // [0][]    Rotor # 10
        [3,8,1,0,5,9,2,7,6,4],  // [1][]            20
        [4,0,8,6,1,5,3,2,9,7],  // [2][]            30
        [3,9,8,0,5,2,6,1,7,4],  // [3][]            40
        [6,4,9,7,1,3,5,2,8,0]   // [4][]            50
        ];

    // Position of a cipher or control rotor is its clockwise displacement from having
    // 'A' on top.  For an index rotor this is its counter clockwise displacement from 
    // zero on top.
    
    constructor() {
		this.pos = 0;           								// Position of the rotor.
    	this.reversed = false;  								// Is this rotor reversed?
	}

/***********************************************************************
 * rotCW() - Rotate a cipher or control rotor clockwise.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * SUMMARY:
 *  rotate clockwise one position of a cipher or control rotor.  Note
 *  that these rotors are labeled clockwise increasing, except when
 *  reversed.
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96 	    1.00    RSP 	First release.
 * 8 Apr 2024		1.00	SRM		First Javascript release
 *      
 ***********************************************************************/
    rotCW() {
                    
        if (this.reversed) {             // Reversed rotors increase counter clockwise.
            this.pos = (this.pos + 1) % 26;
            }
        else {
            this.pos = (this.pos - 1 + 26) % 26;  // Adding 26 guarantees a positive value.
            }
            
        return(this.pos);
        }

/***********************************************************************
 * rotCCW() - Rotate a cipher or control rotor counter-clockwise.
 * Copyright (C) 1996, by Richard Pekelney
 * Javascript code copyright 2024 Stephen R. MacMinn
 * 
 * All Rights Reserved
 *
 * SUMMARY:
 *  rotate counter clockwise one position of a cipher or control rotor. Note
 *  that these rotors are labeled clockwise increasing, except when reversed.
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96 	    1.00    RSP 	First release.
 * 8 Apr 2024		1.00	SRM		First Javascript release
 *      
 ***********************************************************************/
   rotCCW() {

        if (this.reversed) {                 		// Reversed rotors increase counter clockwise.
            this.pos = (this.pos - 1 + 26) % 26; 	// Adding 26 guarantees a positive value.
            }
        else {                          			// Normal rotor increase clockwise.
            this.pos = (this.pos + 1) % 26;
            }
            
        return(this.pos);
        }

/***********************************************************************
 * reverse() - Reverse a rotor.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * SUMMARY:
 * Reversed rotors can be thought of as upside and backwards rotors.  This routine
 * just provides an interface to the reversed variable.
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96 	    1.00    RSP 	First release.
 * 8 Apr 2024		1.00	SRM		First javascript release.
 *      
 ***********************************************************************/
    reverse() {

        this.reversed = true;
        return;
    } 

	static getActiveRotorSetName() {
		return Rotor._activeRotorSetName;
	} 
	
	static installRotorSet(name, wiring) {
		if(name == Rotor.DEFAULT_ROTOR_SET_NAME) {
			Rotor.WIRING = [...Rotor.defaultWIRING];
			Rotor._activeRotorSetName = Rotor.DEFAULT_ROTOR_SET_NAME;
			
		} else {
			Rotor.WIRING = [...wiring];
			Rotor._activeRotorSetName = name;
		}
		return [true, "Rotor set " + name + " installed."];
	} // end of installRotorSet.

} // end of Rotor
