/***********************************************************************
 * RotorCage - contains the workings of the ECM rotors.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 * 
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * SUMMARY:
 * This object contains the guts of the ECM algorithm.
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96		1.00    	RSP 	First release.
 * 8 Apr 2024	1.00		SRM		First Javascript release.
 *      
 ***********************************************************************/
class RotorCage {
	static ENCRYPT = false; 
    static DECRYPT = true;
    static CSP889 = 0; 
    static CSP2900 = 1; 
    static CSPNONE = 2;
  
  	// CONTROL_INDEX_889, CONTROL_INDEX_2900, and INDEX_MAG
  	// are defined as arrays below because javascript does not support
  	// static arrays in classes.
  	
    // This table has the wiring between the left side of the control rotor 
    // bank to the left side of the index rotor.  This table is for a CSP-889.

    
    static CONTROL_INDEX_889 =  Array(
        9,1,2,3,3,4,4,4,5,5,5,6,6,6,6,7,7,7,7,7,8,8,8,8,8,8); // index
    //  a b c d e f g h i j k l m n o p q r s t u v w x y z      control
    //  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5

    // This table has the wiring between the left side of the control rotor
    // bank to the left side of the index rotor.  This table is for a CSP-2900.
    // On the CSP-2900, P, Q and R are not connected. They are 9 in the table, but 
    // the code below handles the exception.
    
    static CONTROL_INDEX_2900 = new Array(
        9,1,2,3,3,4,4,4,5,5,5,6,6,6,6,9,9,9,7,7,0,0,8,8,8,8);  // index
    //  a b c d e f g h i j k l m n o p q r s t u v w x y z       control

    // This table has the wiring between the right side of the index rotor bank 
    // to the magnets that rotate the cipher rotors.
    
    static INDEX_MAG = new Array(
        1,5,5,4,4,3,3,2,2,1);  // rotor stepping magnet
    //  0 1 2 3 4 5 6 7 8 9 index contacts minus one to match array
/*
construct() {
	
    this.cipherBank  = new CipherRotor[5];
    this.controlBank = new ControlRotor[5];
    this.indexBank   = new IndexRotor[5];


    this.cipherCount = 0; // counter used to detect improperly installed index rotors.
}
*/
/***********************************************************************
 * RotorCage - Constructor of the RotorCage object
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 * 
 *  javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 *
 * SUMMARY:
 * This object contains the guts of the ECM algorithm.
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    	RSP 	First release.
 * 8 Apr 2024	1.00		SRM		First Javascript release.
 *      
 ***********************************************************************/
constructor(cipherSet, controlSet, indexSet) {
    var cipherNum, controlNum, indexNum;
    this.cipherBank  = new Array(5);
    this.controlBank = new Array(5);
    this.indexBank   = new Array(5);
	    
    this.cipherCount = 0; // counter used to detect improperly installed index rotors.

        // The passed strings contain the order and orientation of the rotors.
        for (let i = 0; i < 5; i++) {
            cipherNum  = cipherSet.charAt(i * 2)  - '0';  // zero
            controlNum = controlSet.charAt(i * 2) - '0';
            indexNum   = indexSet.charAt(i * 2)   - '0';
            
            // Check for out of bounds rotor numbers
            if ((cipherNum < 0) || (cipherNum > (Rotor.WIRING.length - 1)) ) {
                cipherNum = 0;
                }
            if ((controlNum < 0) || (controlNum > (Rotor.WIRING.length - 1)) ) {
                controlNum = 0;
                }
            if ((indexNum < 0) || (indexNum > (Rotor.INDEXWIRING.length -1)) ) {
                indexNum = 0;
                }
            
            // Create the rotor.
            this.cipherBank[i] = new CipherRotor(cipherNum); 
            if(cipherSet.charAt(i * 2 + 1) == 'R')
                this.cipherBank[i].reverse();    
            this.controlBank[i] = new ControlRotor(controlNum);
            if(controlSet.charAt(i * 2 + 1) == 'R')
                this.controlBank[i].reverse();   
            this.indexBank[i] = new IndexRotor(indexNum);
            if(indexSet.charAt(i * 2 + 1) == 'R')
                this.indexBank[i].reverse(); 
        }
        return;
    }

/***********************************************************************
 * zeroize - Positions the cipher and control rotors with letter 'O' on top.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    zeroize() {
        this.setCipherBankPos("OOOOO");
        this.setControlBankPos("OOOOO");
        return;
	}

/***********************************************************************
 * setCipherBankPos - Sets the cipher rotors position.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * REVISION HISTORY:
 *
 *   Date   	Version 	By  	Purpose of Revision
 * -------- 	------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    	RSP 	First release.
 * 8 Apr 2024	1.00		SRM		First Javascript release.
 *    
 ***********************************************************************/
    setCipherBankPos(posString) {
        var i;
    
        for(i = 0; i < 5; i++) {
        
            // if the first or last rotor changes clear the cipherCount
            if ( (i == 0) || (i == 4) ) {
                if (this.cipherBank[i].pos != posString.charAt(i).charCodeAt(0) - 'A'.charCodeAt(0)) {
                    this.cipherCount = 0;
                    }
                }
                
            // now update the cipher bank
            this.cipherBank[i].pos = posString.charAt(i).charCodeAt(0) - 'A'.charCodeAt(0);
            
            }
        return;
    }
        
/***********************************************************************
 * setControlBankPos - Sets the control rotors position.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    setControlBankPos(posString) {
        var i;
    
        for(i = 0; i < 5; i++) {
            this.controlBank[i].pos = posString.charAt(i).charCodeAt(0) - 'A'.charCodeAt(0);
            }
        return;
	}
        
/***********************************************************************
 * setIndexBankPos - Sets the index rotors position.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    setIndexBankPos(posString) {
    
        for(let i = 0; i < 5; i++) {
            this.indexBank[i].pos = posString.charAt(i).charCodeAt(0) - '0'.charCodeAt(0);
        }
        return;
	}

/***********************************************************************
 * controlBankUpdate - Updates the Control rotor positions between cycles.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * SUMMARY:
 * Control rotors are updated in a "water meter" movement,
 * control rotor 3 first rotates once per cycle, then rotor 4 rotates each time rotor 3
 * moves from the O position, and 2 move least often, rotating once each time rotor 4 
 * moves from the O position.
 *
 * Note that the array indexes are numbered one less than the rotor position names.
 * i.e. [2] is rotor 3 and always rotates, [3] is the 4th rotor, and the 2nd rotor,
 * [1] rotates last.
 *
 * REVISION HISTORY:
 *
 *   Date   Version 	By  	Purpose of Revision
 * -------- ------- 	--- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 12 Feb 98    1.01    RSP 	Slow rotor now only moves when the medium rotor moves.
 *                          	It was incorrectly moving each time medium rotor was O.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    controlBankUpdate() {
		
		/* medium rotor moves */
		
        if (this.controlBank[2].pos == 'O'.charCodeAt(0) - 'A'.charCodeAt(0)) {
			
			/* slow rotor moves */
			
            if (this.controlBank[3].pos == 'O'.charCodeAt(0) - 'A'.charCodeAt(0)) {
                this.controlBank[1].rotCW();
                }
            this.controlBank[3].rotCW();
            }
            
        /* fast rotor always moves */
        
        this.controlBank[2].rotCW();
        return;
	}

/***********************************************************************
 * CipherBankUpdate - Updates the cipher rotor positions between cycles.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * SUMMARY:
 * cipherBankUpdate passes 4 (or 6 for CSP 2900) currents through the control bank using
 * controlBankPath(),
 * then through the wiring from the control bank to the index bank using the CONTROL_INDEX[]
 * table,
 * then through the index bank using indexBankPath(),
 * and finally through the wiring between the index bank and the and the magnets that rotate
 * the cipher rotors using the INDEX_MAG[] table.
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    cipherBankUpdate(machine) {
        var move = new Array(5);
        let i, j, k;

        for (i = 0 ; i < 5 ; i++)
            move[i] = false;

        // The movements are stored in move[] because more than one of the paths through
        // the control and index banks can connect with a single cipher rotor magnet at the
        // same time.  Using the move[] array allows the program to be sequential even though
        // the machine is concurrent and thereby avoid extra motions of the rotor.
        if (machine == CSP889) {
            for (j = 'F'.charCodeAt(0) - 'A'.charCodeAt(0) ; j <= 'I'.charCodeAt(0) - 'A'.charCodeAt(0) ; j++) {
                let cbp = this.controlBankPath(j);
				let x = this.indexBankPath(RotorCage.CONTROL_INDEX_889[this.controlBankPath(j)]);

                move[RotorCage.INDEX_MAG[this.indexBankPath(RotorCage.CONTROL_INDEX_889[this.controlBankPath(j)])]-1] = true;
                }
                
            // Between 1 and 4 cipher rotors will rotate.
            for (i = 0 ; i < 5 ; i++) {
                if (move[i]) {
                    this.cipherBank[i].rotCW();
                    // clear the cipher rotor movement counter if the first or last rotor turn.
                    if (i == 0 || i == 4) {
                        this.cipherCount = 0;
                        }
                    }
                }
            }
        else {  // This is a CSP-2900, there are three changes.
            //1 Six contacts are on instead of four.
            for (j = 'D'.charCodeAt(0) - 'A'.charCodeAt(0) ; j <= 'I'.charCodeAt(0) - 'A'.charCodeAt(0) ; j++) {  
                //2 The control/index wiring is changed and contacts P, Q and R are not connected.
                k = this.controlBankPath(j); 
                if ( (k == 'P'.charCodeAt(0) - 'A'.charCodeAt(0)) || (k == 'Q'.charCodeAt(0) - 'A'.charCodeAt(0)) || (k == 'R'.charCodeAt(0) - 'A'.charCodeAt(0)) ) {
                    continue;  // Skip contacts P, Q and R since they are not connected.
                    }
                move[RotorCage.INDEX_MAG[this.indexBankPath(RotorCage.CONTROL_INDEX_2900[k])]-1] = true;
                }
            // Between 1 and 4 cipher rotors will rotate.
            //3 In a 2900 rotors 2 and 4 ( array index 1 and 3) rotate backwards.
            if (move[0]) {
                this.cipherBank[0].rotCW();
                this.cipherCount = 0;
                }
            if (move[1]) {
                this.cipherBank[1].rotCCW();
                }
            if (move[2]) {
        this.cipherBank[2].rotCW();
        }
            if (move[3]) {
                this.cipherBank[3].rotCCW();
                }
            if (move[4]) {
                this.cipherBank[4].rotCW();
                this.cipherCount = 0;
                }
            }
        return;
	}

/***********************************************************************
 * cipherBankPath - Passes a current through 5 cipher rotors.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 8
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 3 Sep 98     1.01    RSP 	Path reversed between encrypt and decrypt
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    cipherBankPath(direction , pos) {
        var c;
        var rotNum;

        c = pos;
        if (direction == ENCRYPT) {
            // encrypt from left to right
            for (rotNum = 0 ; rotNum <= 4 ; rotNum++)
                c = this.cipherBank[rotNum].cipherEncPath(c);
            }
        else {
            // decrypt from right to left
            for (rotNum = 4 ; rotNum >= 0 ; rotNum--)
                c = this.cipherBank[rotNum].cipherDecPath(c);
            }
        return(c);
	}

/***********************************************************************
 * controlBankPath - Passes a current through 5 control rotors.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    controlBankPath(pos) {
        var c;
        var rotNum;

        // Control rotor bank is always read right to left
        c = pos;
        for (rotNum = 4 ; rotNum >= 0 ; rotNum--)
            c = this.controlBank[rotNum].controlPath(c);
        return(c);
 	}

/***********************************************************************
 * indexBankPath - Passes a current through 5 index rotors.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    indexBankPath(pos) {
        var c;
        var rotNum;

        // Index rotor bank is always read left to right
        c = pos;
        for (rotNum = 0 ; rotNum <= 4 ; rotNum++)
            c = this.indexBank[rotNum].indexPath(c);
        return(c);
	}

/***********************************************************************
 * cipherBankPosToString - Creates a string of cipher rotor positions for display.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 * 
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    cipherBankPosToString() {
        var c = new Array(5);
        var rotNum;
        
        for (rotNum = 0 ; rotNum < 5 ; rotNum++) {
            let code  = (this.cipherBank[rotNum].pos + 'A'.charCodeAt(0));
            c[rotNum] = String.fromCharCode(code)
        }
		return c.join("");
    }
        
/***********************************************************************
 * controlBankPosToString - Creates a string of control rotor positions for display.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    controlBankPosToString() {
        var c = new Array(5);
        var rotNum;
        
        for (rotNum = 0 ; rotNum < 5 ; rotNum++) {
            let code  = (this.controlBank[rotNum].pos + 'A'.charCodeAt(0));
            c[rotNum] = String.fromCharCode(code)
            }
        return c.join("");
    }

/***********************************************************************
 * indexBankPosToString - Creates a string of index rotor positions for display.
 * Copyright (C) 1996, by Richard Pekelney
 * All Rights Reserved
 *
 * javascript version Copyright (C) 2024 by Stephen MacMinn
 *
 * SUMMARY:
 * Creates a string of rotor positions, note the real rotors are numbered 10-19,
 * 20-29, 30-39, 40-49, 50-59, but for now this program uses a single digit index position.
 *
 * REVISION HISTORY:
 *
 *   Date   	Version By  	Purpose of Revision
 * -------- 	------- --- 	--------------------------------------
 * 8 Oct 96     1.00    RSP 	First release.
 * 8 Apr 2024	1.00	SRM		First Javascript release.
 *      
 ***********************************************************************/
    indexBankPosToString() {
        var c = new Array(5);
        var rotNum;
        for (rotNum = 0 ; rotNum < 5 ; rotNum++) {
            let code = this.indexBank[rotNum].pos + '0'.charCodeAt(0);
            c[rotNum] = String.fromCharCode(code);
        }
        return c.join("");
    }      
}   // end of RotorCage

